﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HEAL.Attic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Parameters;

namespace HeuristicLab.Analysis.FitnessLandscape {

  [Item("QAPPermutationFitnessDistanceCorrelationAnalyzer", "An operator that analyzes the correlation between fitness and distance to the best know solution for permutation encoding")]
  [StorableType("31823F7B-55FB-4D57-80EE-89F4D90B6412")]
  public class QAPPermutationFitnessDistanceCorrelationAnalyzer : FitnessDistanceCorrelationAnalyzer, IPermutationOperator {

    #region Parameters
    public ScopeTreeLookupParameter<Permutation> PermutationParameter {
      get { return (ScopeTreeLookupParameter<Permutation>)Parameters["Permutation"]; }
    }
    public LookupParameter<Permutation> BestKnownSolution {
      get { return (LookupParameter<Permutation>)Parameters["BestKnownSolution"]; }
    }
    public ILookupParameter<DoubleMatrix> WeightsParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["Weights"]; }
    }
    public ILookupParameter<DoubleMatrix> DistancesParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["Distances"]; }
    }
    #endregion

    [StorableConstructor]
    protected QAPPermutationFitnessDistanceCorrelationAnalyzer(StorableConstructorFlag _) : base(_) { }
    protected QAPPermutationFitnessDistanceCorrelationAnalyzer(QAPPermutationFitnessDistanceCorrelationAnalyzer original, Cloner cloner) : base(original, cloner) { }

    public QAPPermutationFitnessDistanceCorrelationAnalyzer() {
      Parameters.Add(new ScopeTreeLookupParameter<Permutation>("Permutation", "The permutation encoded solution"));
      Parameters.Add(new LookupParameter<Permutation>("BestKnownSolution", "The best known solution"));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Weights", "The weights matrix."));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Distances", "The distances matrix."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new QAPPermutationFitnessDistanceCorrelationAnalyzer(this, cloner);
    }

    public static double Distance(Permutation a, Permutation b, DoubleMatrix weights, DoubleMatrix distances) {
      Dictionary<string, int> alleles = new Dictionary<string, int>(a.Length * a.Length);
      int distance = 0;
      for (int x = 0; x < a.Length; x++) {
        for (int y = 0; y < a.Length; y++) {
          string alleleA = weights[x, y].ToString() + ">" + distances[a[x], a[y]].ToString();
          string alleleB = weights[x, y].ToString() + ">" + distances[b[x], b[y]].ToString();
          if (alleleA == alleleB) continue;

          int countA = 1, countB = -1;
          if (alleles.ContainsKey(alleleA)) countA += alleles[alleleA];
          if (alleles.ContainsKey(alleleB)) countB += alleles[alleleB];

          if (countA <= 0) distance--; // we've found in A an allele that was present in B
          else distance++; // we've found in A a new allele
          alleles[alleleA] = countA;

          if (countB >= 0) distance--; // we've found in B an allele that was present in A
          else distance++; // we've found in B a new allele
          alleles[alleleB] = countB;
        }
      }
      return distance;
    }

    protected override IEnumerable<double> GetDistancesToBestKnownSolution() {
      if (PermutationParameter.ActualName == null)
        return new double[0];
      Permutation bestKnownValue = BestKnownSolution.ActualValue;
      if (bestKnownValue == null)
        return PermutationParameter.ActualValue.Select(v => 0d);
      return PermutationParameter.ActualValue.Select(v => Distance(v, bestKnownValue, WeightsParameter.ActualValue, DistancesParameter.ActualValue));
    }
  }
}